from argparse import Namespace
from redneck import config, resolver, builder

import logging as log
from pathlib import Path

def build(args: Namespace):
    groups_split = []
    if args.groups != None:
        groups_split = str(args.groups).split(",")
        log.debug(f"Also using group(s) {groups_split}")

    proj = config.scan_project(Path("."))
    if proj == None:
        return    

    used_mods: list[config.ModDecl] = []

    for id, decl in proj.decls.items():
        if decl.group == "base" or decl.group in groups_split:
            used_mods.append(decl)

    log.info(f"Enumerated {len(proj.decls)} mods, resolving {len(used_mods)} mods")

    mods = resolver.resolve_mods(used_mods)
    final_build = builder.build_project(args.builder, builder.ResolvedProject(proj.meta, mods))

    if final_build == None:
        log.error("Build failed")
        return

    log.info(f"Built {final_build}")


from redneck import config, resolver, builder

from argparse import Namespace
from pathlib import Path
import logging as log

def check(args: Namespace):
    proj = config.scan_project(Path("."))
    if proj == None:
        return

    warns = resolver.health_check(proj)
    log.warning(f"{len(warns)} mods have warnings.")

    for id, warn in warns.items():
        decl = proj.decls[id]

        log.warning(f"{id} ({decl._from_file}:{decl._line}:{decl._col}):")
        for i in warn:
            log.warning(f" * {i}")


from redneck import builder, plugins, config, resolver

import requests
import glob
import json

from zipfile import ZipFile
from pathlib import Path
from pydantic import BaseModel
from requests_cache import CachedSession
from typing import Literal, override

class ModrinthPlugin(plugins.RedneckPlugin):
    def __init__(self):
        super().__init__()
        plugins.register_decl("modrinth", ModrinthModDecl, ModrinthResolver())
        plugins.register_builder("mrpack", ModrinthBuilder(), ModrinthBuilderConfig)

class ModrinthModDecl(config.ModDecl):
    load: Literal["modrinth"]
    project: str
    version: str
    file: int = 0

class ModrinthResolver(resolver.ModResolver[ModrinthModDecl]):
    session: requests.Session

    def __init__(self) -> None:
        super().__init__()

    @override
    def resolve(self, decl: ModrinthModDecl) -> resolver.ResolvedMod:
        try:
            self.session
        except AttributeError:
            self.session = CachedSession(".redneck/rinth_cache", backend="filesystem", cache_control=True)

        info = self.session.get(f"https://api.modrinth.com/v2/project/{decl.project}").json()
        ver = self.session.get(f"https://api.modrinth.com/v2/project/{decl.project}/version/{decl.version}").json()

        file = ver["files"][decl.file]

        return resolver.ResolvedMod(file["url"], {
            "name": info["title"],
            "sha512": file["hashes"]["sha512"],
            "sha1": file["hashes"]["sha1"],
            "filename": file["filename"],
            "size": file["size"],
        })

    def health_check(self, proj: config.RedneckProject, decl: ModrinthModDecl) -> list[str]:
        warnings: list[str] = []

        # info = self.session.get(f"https://api.modrinth.com/v2/project/{decl.project}").json()
        ver = self.session.get(f"https://api.modrinth.com/v2/project/{decl.project}/version/{decl.version}").json()

        if not proj.meta.loader in ver["loaders"]:
            warnings.append(f"Unsupported loader. This mod version supports: {ver["loaders"]}")
        if proj.meta.minecraft not in ver["game_versions"]:
            warnings.append(f"Unsupported game version. This mod version supports: {ver["game_versions"]}")

        return warnings

class ModrinthBuilderConfig(BaseModel):
    include_files: list[str] = []

class ModrinthBuilder(builder.ModpackBuilder[ModrinthBuilderConfig]):
    @override
    def build(self, proj: builder.ResolvedProject, root: Path, options: ModrinthBuilderConfig) -> Path:
        build_path = root / ".redneck" / "build" / f"{proj.meta.pack.id}.mrpack"

        index = {
            "game": "minecraft",
            "name": proj.meta.pack.name,
            "versionId": proj.meta.pack.version,
            "formatVersion": 1,
            "dependencies": {
                "minecraft": proj.meta.minecraft,
            },
            "files": []
        }
        index["dependencies"][proj.meta.loader] = proj.meta.loader_version

        for mod in proj.mods:
            file_decl = {
                "path": f"mods/{mod.extra_info["filename"]}",
                "hashes": {
                    "sha1": mod.extra_info["sha1"],
                    "sha512": mod.extra_info["sha512"]
                },
                "downloads": [mod.url],
                "fileSize": mod.extra_info["size"]
            }

            index["files"].append(file_decl)

        includes: set[str] = set()

        for glop in options.include_files:
            includes.update(glob.glob(glop, root_dir=root, recursive=True, include_hidden=True))

        with ZipFile(build_path, "w") as zip:
            zip.writestr("modrinth.index.json", json.dumps(index, separators=(",", ":")))

            for i in includes:
                if (root / i).is_dir():
                    continue

                zip.write(i, f"overrides/{i}")

        return build_path


import argparse
import logging as log
import inspect
from typing import Any

from pydantic import BaseModel
from redneck import config, resolver, builder
from pkgutil import iter_modules
from importlib import import_module

class RedneckPlugin():
    def __init__(self):
        pass

    def add_subparser(self, parser: argparse.ArgumentParser):
        pass

_instances: list[RedneckPlugin] = []

def plugins():
    log.info(f"{len(_instances)} plugins:")
    for i in _instances:
        log.info(f"- {type(i)}")

def load_plugins():
    """
    Shamelessly stolen from beets, my favorite music database manager.
    """
    if _instances: 
        return

    import redplug
    discovered_plugins: dict[str, RedneckPlugin] = {}

    for module in iter_modules(redplug.__path__, redplug.__name__ + "."):
        try:
            namespace = import_module(module.name)
            for obj in reversed(namespace.__dict__.values()):
                if (
                    inspect.isclass(obj)
                    and issubclass(obj, RedneckPlugin)
                    and obj != RedneckPlugin
                    and not inspect.isabstract(obj)
                    # Only consider this plugin's module or submodules to avoid
                    # conflicts when plugins import other plugin classes
                    and (
                        obj.__module__ == namespace.__name__
                        or obj.__module__.startswith(f"{namespace.__name__}.")
                    )
                ):
                    plugin = obj()
                    discovered_plugins[module.name] = plugin
                    _instances.append(plugin)

                    break
        except Exception as e:
            log.error(f"Failed to load plugin {module.name}: {e}");
    log.debug(f"Loaded {len(discovered_plugins)} plugins: {list(discovered_plugins.keys())}")

def register_decl(id: str, decl: type[resolver.T], resl: resolver.ModResolver[resolver.T]):
    config._decls.append(decl) 
    # I am sorry, lord. This simply is the easiest way.
    resolver._resolvers[id] = resl

def register_builder(id: str, bldr: builder.ModpackBuilder[builder.T], conf: type[BaseModel] | None = None):
    builder._builders[id] = bldr
    if conf != None:
        builder._configs[id] = conf


from pathlib import Path
from typing import override
from redneck import plugins, builder

class ModlistPlugin(plugins.RedneckPlugin):
    def __init__(self):
        super().__init__()
        plugins.register_builder("modlist", ModlistBuilder())

class ModlistBuilder(builder.ModpackBuilder):
    @override
    def build(self, proj: builder.ResolvedProject, root: Path, options: None) -> Path:
        out = root / ".redneck" / "build" / "modlist.html"

        with open(out, "w") as file:
            file.write("<ul>\n")
            for mod in proj.mods:
                file.write(f"<li><a href=\"{mod.url}\">{mod.extra_info["name"] if "name" in mod.extra_info else mod.url}</a></li>\n")
            file.write("</ul>")

        return out


from redneck import config, resolver

import logging as log
from traceback import format_exc
from typing import Generic, TypeVar
from pydantic import BaseModel
from pathlib import Path

T = TypeVar("T", bound=BaseModel)

class ResolvedProject():
    meta: config.ModpackMeta
    mods: list[resolver.ResolvedMod]

    def __init__(self, meta: config.ModpackMeta, mods: list[resolver.ResolvedMod]) -> None:
        self.meta = meta
        self.mods = mods

class ModpackBuilder(Generic[T]):
    def build(self, proj: ResolvedProject, root: Path, options: T | None) -> Path:
        raise NotImplementedError

_builders: dict[str, ModpackBuilder] = {}
_configs: dict[str, type[BaseModel]] = {}

def build_project(id: str, proj: ResolvedProject) -> Path | None:
    root = Path(".")
    builds = root / ".redneck" / "build"
    if not builds.exists():
        builds.mkdir(parents=True)

    config = None

    if id in _configs.keys():
        raw = proj.meta.builders.get(id, {})
        try:
            config = _configs[id].model_validate(raw)
        except Exception as e:
            log.error(f"In pack.yml:\n{e}")
            return None

    try:
        return _builders[id].build(proj, root, config)
    except Exception as e:
        log.exception(e)


import sys
import coloredlogs
import logging as log
import argparse
from importlib.metadata import version
from pathlib import Path
from redneck.commands import build, groups, check
from redneck import plugins, config, resolver, builder

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--verbose", help="Enables debug output for the logger", action="store_true")

    plugins.load_plugins()

    subparser = parser.add_subparsers(dest="action")
    subparser.required = True

    build_command = subparser.add_parser("build", help="Builds the project.")
    build_command.add_argument("-g", "--groups", help="The mod groups to include in the build.")
    build_command.add_argument("builder", help="The builder to use for building the modpack.")

    groups_command = subparser.add_parser("groups", help="Lists the groups defined in the mods.")
    groups_command.add_argument("-f", "--full", help="Lists groups, as well as mod ids and their source files.", action="store_true")

    check_command = subparser.add_parser("check", help="Checks mod decls for validity.")

    plugins_command = subparser.add_parser("plugins", help="Lists the plugins loaded during startup.")

    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    # Change the color because I can't view black on black 
    coloredlogs.DEFAULT_FIELD_STYLES["levelname"]["color"] = "magenta"
    coloredlogs.install(fmt="[%(levelname)-7s] %(message)s", level="DEBUG" if args.verbose else "INFO")
    log.info(f"Redneck v{version("redneck")}")
    
    match args.action:
        case "build":
            build.build(args)
        case "groups":
            groups.groups(args)
        case "check":
            check.check(args)
        case "plugins":
            plugins.plugins()

if __name__ == "__main__":
    main()